@@ -30,8 +30,10 @@ module Language.Haskell.TH.Desugar.Sweeten (
   conToTH, foreignToTH, pragmaToTH, ruleBndrToTH,
   clauseToTH, tvbToTH, cxtToTH, predToTH, derivClauseToTH,
 #if __GLASGOW_HASKELL__ >= 801
-  patSynDirToTH
+  patSynDirToTH,
 #endif
+
+  typeArgToTH
   ) where
 
 import Prelude hiding (exp)
@@ -40,6 +42,7 @@ import Control.Arrow
 import Language.Haskell.TH hiding (cxt)
 
 import Language.Haskell.TH.Desugar.AST
+import Language.Haskell.TH.Desugar.Core (DTypeArg(..))
 import Language.Haskell.TH.Desugar.Util
 
 import Data.Maybe ( maybeToList, mapMaybe )
@@ -70,13 +73,13 @@ matchToTH :: DMatch -> Match
 matchToTH (DMatch pat exp) = Match (patToTH pat) (NormalB (expToTH exp)) []
 
 patToTH :: DPat -> Pat
-patToTH (DLitPa lit)    = LitP lit
-patToTH (DVarPa n)      = VarP n
-patToTH (DConPa n pats) = ConP n (map patToTH pats)
-patToTH (DTildePa pat)  = TildeP (patToTH pat)
-patToTH (DBangPa pat)   = BangP (patToTH pat)
-patToTH (DSigPa pat ty) = SigP (patToTH pat) (typeToTH ty)
-patToTH DWildPa         = WildP
+patToTH (DLitP lit)    = LitP lit
+patToTH (DVarP n)      = VarP n
+patToTH (DConP n pats) = ConP n (map patToTH pats)
+patToTH (DTildeP pat)  = TildeP (patToTH pat)
+patToTH (DBangP pat)   = BangP (patToTH pat)
+patToTH (DSigP pat ty) = SigP (patToTH pat) (typeToTH ty)
+patToTH DWildP         = WildP
 
 decsToTH :: [DDec] -> [Dec]
 decsToTH = concatMap decToTH
@@ -104,12 +107,18 @@ decToTH (DDataD Newtype cxt n tvbs _mk [con] derivings) =
 decToTH (DTySynD n tvbs ty) = [TySynD n (map tvbToTH tvbs) (typeToTH ty)]
 decToTH (DClassD cxt n tvbs fds decs) =
   [ClassD (cxtToTH cxt) n (map tvbToTH tvbs) fds (decsToTH decs)]
-#if __GLASGOW_HASKELL__ >= 711
-decToTH (DInstanceD over cxt ty decs) =
-  [InstanceD over (cxtToTH cxt) (typeToTH ty) (decsToTH decs)]
+decToTH (DInstanceD over mtvbs _cxt _ty decs) =
+  [instanceDToTH over cxt' ty' decs]
+  where
+  (cxt', ty') = case mtvbs of
+                  Nothing    -> (_cxt, _ty)
+                  Just _tvbs ->
+#if __GLASGOW_HASKELL__ < 800 || __GLASGOW_HASKELL__ >= 802
+                                ([], DForallT _tvbs _cxt _ty)
 #else
-decToTH (DInstanceD _ cxt ty decs) =
-  [InstanceD (cxtToTH cxt) (typeToTH ty) (decsToTH decs)]
+                                -- See #117
+                                error $ "Explicit foralls in instance declarations "
+                                     ++ "are broken on GHC 8.0."
 #endif
 decToTH (DForeignD f) = [ForeignD (foreignToTH f)]
 #if __GLASGOW_HASKELL__ > 710
@@ -125,53 +134,46 @@ decToTH (DDataFamilyD n tvbs mk) =
 #else
   [FamilyD DataFam n (map tvbToTH tvbs) (fmap typeToTH mk)]
 #endif
-decToTH (DDataInstD Data cxt n tys _mk cons derivings) =
-#if __GLASGOW_HASKELL__ > 710
-  [DataInstD (cxtToTH cxt) n (map typeToTH tys) (fmap typeToTH _mk) (map conToTH cons)
-             (concatMap derivClauseToTH derivings)]
+decToTH (DDataInstD nd cxt mtvbs lhs mk cons derivings) =
+  let ndc = case (nd, cons) of
+              (Newtype, [con]) -> DNewtypeCon con
+              (Newtype, _)     -> error "Newtype that doesn't have only one constructor"
+              (Data,    _)     -> DDataCons cons
+  in dataInstDecToTH ndc cxt mtvbs lhs mk derivings
+#if __GLASGOW_HASKELL__ >= 807
+decToTH (DTySynInstD eqn) = [TySynInstD (snd $ tySynEqnToTH eqn)]
 #else
-  [DataInstD (cxtToTH cxt) n (map typeToTH tys) (map conToTH cons)
-             (map derivingToTH derivings)]
+decToTH (DTySynInstD eqn) =
+  let (n, eqn') = tySynEqnToTH eqn in
+  [TySynInstD n eqn']
 #endif
-decToTH (DDataInstD Newtype cxt n tys _mk [con] derivings) =
-#if __GLASGOW_HASKELL__ > 710
-  [NewtypeInstD (cxtToTH cxt) n (map typeToTH tys) (fmap typeToTH _mk) (conToTH con)
-                (concatMap derivClauseToTH derivings)]
-#else
-  [NewtypeInstD (cxtToTH cxt) n (map typeToTH tys) (conToTH con)
-                (map derivingToTH derivings)]
-#endif
-#if __GLASGOW_HASKELL__ < 707
-decToTH (DTySynInstD n eqn) = [tySynEqnToTHDec n eqn]
-decToTH (DClosedTypeFamilyD (DTypeFamilyHead n tvbs frs _ann) eqns) =
-  (FamilyD TypeFam n (map tvbToTH tvbs) (frsToTH frs)) :
-  (map (tySynEqnToTHDec n) eqns)
-decToTH (DRoleAnnotD {}) = []
-#else
-decToTH (DTySynInstD n eqn) = [TySynInstD n (tySynEqnToTH eqn)]
 #if __GLASGOW_HASKELL__ > 710
 decToTH (DClosedTypeFamilyD (DTypeFamilyHead n tvbs frs ann) eqns) =
   [ClosedTypeFamilyD (TypeFamilyHead n (map tvbToTH tvbs) (frsToTH frs) ann)
-                     (map tySynEqnToTH eqns)
+                     (map (snd . tySynEqnToTH) eqns)
   ]
 #else
 decToTH (DClosedTypeFamilyD (DTypeFamilyHead n tvbs frs _ann) eqns) =
-  [ClosedTypeFamilyD n (map tvbToTH tvbs) (frsToTH frs) (map tySynEqnToTH eqns)]
+  [ClosedTypeFamilyD n (map tvbToTH tvbs) (frsToTH frs) (map (snd . tySynEqnToTH) eqns)]
 #endif
 decToTH (DRoleAnnotD n roles) = [RoleAnnotD n roles]
+decToTH (DStandaloneDerivD mds mtvbs _cxt _ty) =
+  [standaloneDerivDToTH mds cxt' ty']
+  where
+  (cxt', ty') = case mtvbs of
+                  Nothing    -> (_cxt, _ty)
+                  Just _tvbs ->
+#if __GLASGOW_HASKELL__ < 710 || __GLASGOW_HASKELL__ >= 802
+                                ([], DForallT _tvbs _cxt _ty)
+#else
+                                -- See #117
+                                error $ "Explicit foralls in standalone deriving declarations "
+                                     ++ "are broken on GHC 7.10 and 8.0."
 #endif
 #if __GLASGOW_HASKELL__ < 709
-decToTH (DStandaloneDerivD {}) =
-  error "Standalone deriving supported only in GHC 7.10+"
 decToTH (DDefaultSigD {})      =
   error "Default method signatures supported only in GHC 7.10+"
 #else
-decToTH (DStandaloneDerivD _mds cxt ty) =
-  [StandaloneDerivD
-#if __GLASGOW_HASKELL__ >= 801
-    (fmap derivStrategyToTH _mds)
-#endif
-    (cxtToTH cxt) (typeToTH ty)]
 decToTH (DDefaultSigD n ty)        = [DefaultSigD n (typeToTH ty)]
 #endif
 #if __GLASGOW_HASKELL__ >= 801
@@ -186,6 +188,49 @@ decToTH dec
 #endif
 decToTH _ = error "Newtype declaration without exactly 1 constructor."
 
+-- | Indicates whether something is a newtype or data type, bundling its
+-- constructor(s) along with it.
+data DNewOrDataCons
+  = DNewtypeCon DCon
+  | DDataCons   [DCon]
+
+-- | Sweeten a 'DDataInstD'.
+dataInstDecToTH :: DNewOrDataCons -> DCxt -> Maybe [DTyVarBndr] -> DType
+                -> Maybe DKind -> [DDerivClause] -> [Dec]
+dataInstDecToTH ndc cxt _mtvbs lhs _mk derivings =
+  case ndc of
+    DNewtypeCon con ->
+#if __GLASGOW_HASKELL__ >= 807
+      [NewtypeInstD (cxtToTH cxt) (fmap (fmap tvbToTH) _mtvbs) (typeToTH lhs)
+                    (fmap typeToTH _mk) (conToTH con)
+                    (concatMap derivClauseToTH derivings)]
+#elif __GLASGOW_HASKELL__ > 710
+      [NewtypeInstD (cxtToTH cxt) _n _lhs_args (fmap typeToTH _mk) (conToTH con)
+                    (concatMap derivClauseToTH derivings)]
+#else
+      [NewtypeInstD (cxtToTH cxt) _n _lhs_args (conToTH con)
+                    (map derivingToTH derivings)]
+#endif
+
+    DDataCons cons ->
+#if __GLASGOW_HASKELL__ >= 807
+      [DataInstD (cxtToTH cxt) (fmap (fmap tvbToTH) _mtvbs) (typeToTH lhs)
+                 (fmap typeToTH _mk) (map conToTH cons)
+                 (concatMap derivClauseToTH derivings)]
+#elif __GLASGOW_HASKELL__ > 710
+      [DataInstD (cxtToTH cxt) _n _lhs_args (fmap typeToTH _mk) (map conToTH cons)
+                 (concatMap derivClauseToTH derivings)]
+#else
+      [DataInstD (cxtToTH cxt) _n _lhs_args (map conToTH cons)
+                 (map derivingToTH derivings)]
+#endif
+  where
+    _lhs' = typeToTH lhs
+    (_n, _lhs_args) =
+      case unfoldType _lhs' of
+        (ConT n, lhs_args) -> (n, filterTANormals lhs_args)
+        (_, _) -> error $ "Illegal data instance LHS: " ++ pprint _lhs'
+
 #if __GLASGOW_HASKELL__ > 710
 frsToTH :: DFamilyResultSig -> FamilyResultSig
 frsToTH DNoSig          = NoSig
@@ -201,7 +246,7 @@ frsToTH (DTyVarSig (DKindedTV _ k)) = Just (typeToTH k)
 
 #if __GLASGOW_HASKELL__ <= 710
 derivingToTH :: DDerivClause -> Name
-derivingToTH (DDerivClause _ [DConPr nm]) = nm
+derivingToTH (DDerivClause _ [DConT nm]) = nm
 derivingToTH p =
   error ("Template Haskell in GHC < 8.0 only allows simple derivings: " ++ show p)
 #endif
@@ -268,19 +313,41 @@ conToTH (DCon tvbs cxt n fields rty)
     num_univ_tvs = go rty
       where
         go :: DType -> Int
-        go (DForallT {}) = error "`forall` type used in GADT return type"
         go (DAppT t1 t2) = go t1 + go t2
         go (DSigT t _)   = go t
         go (DVarT {})    = 1
         go (DConT {})    = 0
         go DArrowT       = 0
         go (DLitT {})    = 0
-        go DWildCardT    = 0
+        -- These won't show up on pre-8.0 GHCs
+        go (DForallT {})  = error "`forall` type used in GADT return type"
+        go DWildCardT     = 0
+        go (DAppKindT {}) = 0
 
     con' :: Con
     con' = conToTH $ DCon [] [] n fields rty
 #endif
 
+instanceDToTH :: Maybe Overlap -> DCxt -> DType -> [DDec] -> Dec
+instanceDToTH _over cxt ty decs =
+  InstanceD
+#if __GLASGOW_HASKELL__ >= 800
+            _over
+#endif
+            (cxtToTH cxt) (typeToTH ty) (decsToTH decs)
+
+standaloneDerivDToTH :: Maybe DDerivStrategy -> DCxt -> DType -> Dec
+#if __GLASGOW_HASKELL__ >= 710
+standaloneDerivDToTH _mds cxt ty =
+  StandaloneDerivD
+#if __GLASGOW_HASKELL__ >= 802
+                   (fmap derivStrategyToTH _mds)
+#endif
+                   (cxtToTH cxt) (typeToTH ty)
+#else
+standaloneDerivDToTH _ _ _ = error "Standalone deriving supported only in GHC 7.10+"
+#endif
+
 foreignToTH :: DForeign -> Foreign
 foreignToTH (DImportF cc safety str n ty) =
   ImportF cc safety str n (typeToTH ty)
@@ -291,13 +358,15 @@ pragmaToTH (DInlineP n inl rm phases) = Just $ InlineP n inl rm phases
 pragmaToTH (DSpecialiseP n ty m_inl phases) =
   Just $ SpecialiseP n (typeToTH ty) m_inl phases
 pragmaToTH (DSpecialiseInstP ty) = Just $ SpecialiseInstP (typeToTH ty)
-pragmaToTH (DRuleP str rbs lhs rhs phases) =
-  Just $ RuleP str (map ruleBndrToTH rbs) (expToTH lhs) (expToTH rhs) phases
-#if __GLASGOW_HASKELL__ < 707
-pragmaToTH (DAnnP {}) = Nothing
+#if __GLASGOW_HASKELL__ >= 807
+pragmaToTH (DRuleP str mtvbs rbs lhs rhs phases) =
+  Just $ RuleP str (fmap (fmap tvbToTH) mtvbs) (map ruleBndrToTH rbs)
+               (expToTH lhs) (expToTH rhs) phases
 #else
-pragmaToTH (DAnnP target exp) = Just $ AnnP target (expToTH exp)
+pragmaToTH (DRuleP str _ rbs lhs rhs phases) =
+  Just $ RuleP str (map ruleBndrToTH rbs) (expToTH lhs) (expToTH rhs) phases
 #endif
+pragmaToTH (DAnnP target exp) = Just $ AnnP target (expToTH exp)
 #if __GLASGOW_HASKELL__ < 709
 pragmaToTH (DLineP {}) = Nothing
 #else
@@ -313,15 +382,22 @@ ruleBndrToTH :: DRuleBndr -> RuleBndr
 ruleBndrToTH (DRuleVar n) = RuleVar n
 ruleBndrToTH (DTypedRuleVar n ty) = TypedRuleVar n (typeToTH ty)
 
-#if __GLASGOW_HASKELL__ < 707
--- | GHC 7.6.3 doesn't have TySynEqn, so we sweeten to a Dec in GHC 7.6.3;
--- GHC 7.8+ does not use this function
-tySynEqnToTHDec :: Name -> DTySynEqn -> Dec
-tySynEqnToTHDec n (DTySynEqn lhs rhs) =
-  TySynInstD n (map typeToTH lhs) (typeToTH rhs)
+#if __GLASGOW_HASKELL__ >= 807
+-- | It's convenient to also return a 'Name' here, since some call sites make
+-- use of it.
+tySynEqnToTH :: DTySynEqn -> (Name, TySynEqn)
+tySynEqnToTH (DTySynEqn tvbs lhs rhs) =
+  let lhs' = typeToTH lhs in
+  case unfoldType lhs' of
+    (ConT n, _lhs_args) -> (n, TySynEqn (fmap (fmap tvbToTH) tvbs) lhs' (typeToTH rhs))
+    (_, _) -> error $ "Illegal type instance LHS: " ++ pprint lhs'
 #else
-tySynEqnToTH :: DTySynEqn -> TySynEqn
-tySynEqnToTH (DTySynEqn lhs rhs) = TySynEqn (map typeToTH lhs) (typeToTH rhs)
+tySynEqnToTH :: DTySynEqn -> (Name, TySynEqn)
+tySynEqnToTH (DTySynEqn _ lhs rhs) =
+  let lhs' = typeToTH lhs in
+  case unfoldType lhs' of
+    (ConT n, lhs_args) -> (n, TySynEqn (filterTANormals lhs_args) (typeToTH rhs))
+    (_, _) -> error $ "Illegal type instance LHS: " ++ pprint lhs'
 #endif
 
 clauseToTH :: DClause -> Clause
@@ -340,6 +416,13 @@ typeToTH DWildCardT = WildCardT
 #else
 typeToTH DWildCardT = error "Wildcards supported only in GHC 8.0+"
 #endif
+#if __GLASGOW_HASKELL__ >= 807
+typeToTH (DAppKindT t k)        = AppKindT (typeToTH t) (typeToTH k)
+#else
+-- In the event that we're on a version of Template Haskell without support for
+-- kind applications, we will simply drop the applied kind.
+typeToTH (DAppKindT t _)        = typeToTH t
+#endif
 
 tvbToTH :: DTyVarBndr -> TyVarBndr
 tvbToTH (DPlainTV n)           = PlainTV n
@@ -380,35 +463,51 @@ predToTH :: DPred -> Pred
 #if __GLASGOW_HASKELL__ < 709
 predToTH = go []
   where
-    go acc (DAppPr p t) = go (typeToTH t : acc) p
-    go acc (DSigPr p _) = go acc                p  -- this shouldn't happen.
-    go _   (DVarPr _)
-      = error "Template Haskell in GHC <= 7.8 does not support variable constraints."
-    go acc (DConPr n)
+    go acc (DAppT p t) = go (typeToTH t : acc) p
+    -- In the event that we're on a version of Template Haskell without support
+    -- for kind applications, we will simply drop the applied kind.
+    go acc (DAppKindT t _) = go acc t
+    go acc (DSigT p _) = go acc                p  -- this shouldn't happen.
+    go acc (DConT n)
       | nameBase n == "~"
       , [t1, t2] <- acc
       = EqualP t1 t2
       | otherwise
       = ClassP n acc
-    go _ DWildCardPr
+    go _   (DVarT _)
+      = error "Template Haskell in GHC <= 7.8 does not support variable constraints."
+    go _ DWildCardT
       = error "Wildcards supported only in GHC 8.0+"
-    go _ (DForallPr {})
+    go _ (DForallT {})
       = error "Quantified constraints supported only in GHC 8.6+"
+    go _ DArrowT
+      = error "(->) spotted at head of a constraint"
+    go _ (DLitT {})
+      = error "Type-level literal spotted at head of a constraint"
 #else
-predToTH (DAppPr p t) = AppT (predToTH p) (typeToTH t)
-predToTH (DSigPr p k) = SigT (predToTH p) (typeToTH k)
-predToTH (DVarPr n)   = VarT n
-predToTH (DConPr n)   = typeToTH (DConT n)
+predToTH (DAppT p t) = AppT (predToTH p) (typeToTH t)
+predToTH (DSigT p k) = SigT (predToTH p) (typeToTH k)
+predToTH (DVarT n)   = VarT n
+predToTH (DConT n)   = typeToTH (DConT n)
+predToTH DArrowT     = ArrowT
+predToTH (DLitT lit) = LitT lit
 #if __GLASGOW_HASKELL__ > 710
-predToTH DWildCardPr  = WildCardT
+predToTH DWildCardT  = WildCardT
 #else
-predToTH DWildCardPr  = error "Wildcards supported only in GHC 8.0+"
+predToTH DWildCardT  = error "Wildcards supported only in GHC 8.0+"
 #endif
 #if __GLASGOW_HASKELL__ >= 805
-predToTH (DForallPr tvbs cxt p) =
+predToTH (DForallT tvbs cxt p) =
   ForallT (map tvbToTH tvbs) (map predToTH cxt) (predToTH p)
 #else
-predToTH (DForallPr {}) = error "Quantified constraints supported only in GHC 8.6+"
+predToTH (DForallT {}) = error "Quantified constraints supported only in GHC 8.6+"
+#endif
+#if __GLASGOW_HASKELL__ >= 807
+predToTH (DAppKindT p k) = AppKindT (predToTH p) (typeToTH k)
+#else
+-- In the event that we're on a version of Template Haskell without support for
+-- kind applications, we will simply drop the applied kind.
+predToTH (DAppKindT p _) = predToTH p
 #endif
 #endif
 
@@ -430,15 +529,15 @@ tyconToTH n
 #endif
                                   else TupleT deg
   | Just deg <- unboxedTupleNameDegree_maybe n = UnboxedTupleT deg
-#if __GLASGOW_HASKELL__ == 706
-    -- Work around Trac #7667
-  | isTypeKindName n            = StarT
-#endif
 #if __GLASGOW_HASKELL__ >= 801
   | Just deg <- unboxedSumNameDegree_maybe n   = UnboxedSumT deg
 #endif
   | otherwise                   = ConT n
 
+typeArgToTH :: DTypeArg -> TypeArg
+typeArgToTH (DTANormal t) = TANormal (typeToTH t)
+typeArgToTH (DTyArg k)    = TyArg    (typeToTH k)
+
 #if __GLASGOW_HASKELL__ <= 710
 -- | Convert a 'Bang' to a 'Strict'
 bangToStrict :: Bang -> Strict
