@@ -8,7 +8,7 @@ rae@cs.brynmawr.edu
              RankNTypes, TypeFamilies,
              DataKinds, ConstraintKinds, PolyKinds, MultiParamTypeClasses,
              FlexibleInstances, ExistentialQuantification,
-             ScopedTypeVariables, GADTs, ViewPatterns #-}
+             ScopedTypeVariables, GADTs, ViewPatterns, TupleSections #-}
 {-# OPTIONS -fno-warn-incomplete-patterns -fno-warn-overlapping-patterns
             -fno-warn-unused-matches -fno-warn-type-defaults
             -fno-warn-missing-signatures -fno-warn-unused-do-bind
@@ -17,6 +17,7 @@ rae@cs.brynmawr.edu
 #if __GLASGOW_HASKELL__ >= 711
 {-# LANGUAGE PartialTypeSignatures #-}
 {-# LANGUAGE PatternSynonyms #-}
+{-# LANGUAGE TypeApplications #-}
 {-# OPTIONS_GHC -Wno-partial-type-signatures -Wno-redundant-constraints #-}
 #endif
 
@@ -38,9 +39,8 @@ import qualified DsDec
 import qualified Dec
 import Dec ( RecordSel )
 import Language.Haskell.TH.Desugar
-#if __GLASGOW_HASKELL__ >= 707
+import qualified Language.Haskell.TH.Desugar.OSet as OS
 import Language.Haskell.TH.Desugar.Expand  ( expandUnsoundly )
-#endif
 import Language.Haskell.TH
 import qualified Language.Haskell.TH.Syntax as Syn ( lift )
 
@@ -49,13 +49,9 @@ import Control.Monad
 import Control.Applicative
 #endif
 
-import Data.Generics ( geq )
 import Data.Function ( on )
 import qualified Data.Map as M
-import qualified Data.Set as S
-#if __GLASGOW_HASKELL__ >= 707
 import Data.Proxy
-#endif
 
 -- |
 -- Convert a HUnit test suite to a spec.  This can be used to run existing
@@ -86,10 +82,8 @@ tests = test [ "sections" ~: $test1_sections  @=? $(dsSplice test1_sections)
              , "case"     ~: $test8_case      @=? $(dsSplice test8_case)
              , "do"       ~: $test9_do        @=? $(dsSplice test9_do)
              , "comp"     ~: $test10_comp     @=? $(dsSplice test10_comp)
-#if __GLASGOW_HASKELL__ >= 707
              , "parcomp"  ~: $test11_parcomp  @=? $(dsSplice test11_parcomp)
              , "parcomp2" ~: $test12_parcomp2 @=? $(dsSplice test12_parcomp2)
-#endif
              , "sig"      ~: $test13_sig      @=? $(dsSplice test13_sig)
              , "record"   ~: $test14_record   @=? $(dsSplice test14_record)
              , "litp"     ~: $test15_litp     @=? $(dsSplice test15_litp)
@@ -136,6 +130,13 @@ tests = test [ "sections" ~: $test1_sections  @=? $(dsSplice test1_sections)
              , "do_partial_match" ~: $test47_do_partial_match @=? $(dsSplice test47_do_partial_match)
 #if __GLASGOW_HASKELL__ >= 805
              , "quantified_constraints" ~: $test48_quantified_constraints @=? $(dsSplice test48_quantified_constraints)
+#endif
+#if __GLASGOW_HASKELL__ >= 807
+             , "implicit_params" ~: $test49_implicit_params @=? $(dsSplice test49_implicit_params)
+             , "vka"             ~: $test50_vka             @=? $(dsSplice test50_vka)
+#endif
+#if __GLASGOW_HASKELL__ >= 809
+             , "tuple_sections"  ~: $test51_tuple_sections  @=? $(dsSplice test51_tuple_sections)
 #endif
              ]
 
@@ -147,7 +148,6 @@ test_e3a = $test_expand3
 test_e3b = $(test_expand3 >>= dsExp >>= expand >>= return . expToTH)
 test_e4a = $test_expand4
 test_e4b = $(test_expand4 >>= dsExp >>= expand >>= return . expToTH)
-#if __GLASGOW_HASKELL__ >= 707
 test_e5a = $test_expand5
 test_e5b = $(test_expand5 >>= dsExp >>= expand >>= return . expToTH)
 test_e6a = $test_expand6
@@ -161,7 +161,6 @@ test_e8a = $(test_expand8 >>= dsExp >>= expand >>= return . expToTH)
   -- closed type families.
 #endif
 test_e8b = $(test_expand8 >>= dsExp >>= expandUnsoundly >>= return . expToTH)
-#endif
 #if __GLASGOW_HASKELL__ >= 709
 test_e9a = $test_expand9  -- requires GHC #9262
 test_e9b = $(test_expand9 >>= dsExp >>= expand >>= return . expToTH)
@@ -175,7 +174,6 @@ test_expand = and [ hasSameType test35a test35b
                   , hasSameType test36a test36b
                   , hasSameType test_e3a test_e3b
                   , hasSameType test_e4a test_e4b
-#if __GLASGOW_HASKELL__ >= 707
                   , hasSameType test_e5a test_e5b
                   , hasSameType test_e6a test_e6b
                   , hasSameType test_e7a test_e7b
@@ -184,7 +182,6 @@ test_expand = and [ hasSameType test35a test35b
                   , hasSameType test_e8a test_e8a
 #endif
                   , hasSameType test_e8b test_e8b
-#endif
 #if __GLASGOW_HASKELL__ >= 709
                   , hasSameType test_e9a test_e9b
 #endif
@@ -208,15 +205,15 @@ test_mkName = and [ hasSameType (Proxy :: Proxy FuzzSyn) (Proxy :: Proxy Fuzz)
 test_bug8884 :: Bool
 test_bug8884 = $(do info <- reify ''Poly
                     dinfo@(DTyConI (DOpenTypeFamilyD (DTypeFamilyHead _name _tvbs (DKindSig resK) _ann))
-                                   (Just [DTySynInstD _name2 (DTySynEqn lhs _rhs)]))
+                                   (Just [DTySynInstD (DTySynEqn _ lhs _rhs)]))
                       <- dsInfo info
                     let isTypeKind (DConT n) = isTypeKindName n
                         isTypeKind _         = False
                     case (isTypeKind resK, lhs) of
 #if __GLASGOW_HASKELL__ < 709
-                      (True, [DVarT _]) -> [| True |]
+                      (True, _ `DAppT` DVarT _) -> [| True |]
 #else
-                      (True, [DSigT (DVarT _) (DVarT _)]) -> [| True |]
+                      (True, _ `DAppT` DSigT (DVarT _) (DVarT _)) -> [| True |]
 #endif
                       _                                     -> do
                         runIO $ do
@@ -256,7 +253,8 @@ test_local_tyfam_expansion =
        let orig_ty = DConT fam_name
        exp_ty <- withLocalDeclarations
                    (decsToTH [ DOpenTypeFamilyD (DTypeFamilyHead fam_name [] DNoSig Nothing)
-                             , DTySynInstD fam_name (DTySynEqn [] (DConT ''Int)) ])
+                             , DTySynInstD (DTySynEqn Nothing
+                                                      (DConT fam_name) (DConT ''Int)) ])
                    (expandType orig_ty)
        orig_ty `eqTHSplice` exp_ty)
 
@@ -274,8 +272,11 @@ test_stuck_tyfam_expansion =
                                                   (DKindSig (DVarT k))
                                                   Nothing)
                                -- type instance F (x :: ()) = x
-                             , DTySynInstD fam_name
-                                 (DTySynEqn [DSigT (DVarT x) (DConT ''())] (DVarT x))
+                             , DTySynInstD
+                                 (DTySynEqn Nothing
+                                            (DConT fam_name `DAppT`
+                                               DSigT (DVarT x) (DConT ''()))
+                                            (DVarT x))
                              ])
                    (expandType orig_ty)
        orig_ty `eqTHSplice` exp_ty)
@@ -289,6 +290,24 @@ test_t85 =
        expanded_ty <- expandType orig_ty
        expected_ty `eqTHSplice` expanded_ty)
 
+test_t92 :: Bool
+test_t92 =
+  $(do a <- newName "a"
+       f <- newName "f"
+       let t = DForallT [DPlainTV f] [] (DVarT f `DAppT` DVarT a)
+       toposortTyVarsOf [t] `eqTHSplice` [DPlainTV a])
+
+test_t97 :: Bool
+test_t97 =
+  $(do a <- newName "a"
+       k <- newName "k"
+       let orig_ty = DForallT
+                       [DKindedTV a (DConT ''Constant `DAppT` DConT ''Int `DAppT` DVarT k)]
+                       [] (DVarT a)
+           expected_ty = DForallT [DKindedTV a (DVarT k)] [] (DVarT a)
+       expanded_ty <- expandType orig_ty
+       expected_ty `eqTHSplice` expanded_ty)
+
 test_getDataD_kind_sig :: Bool
 test_getDataD_kind_sig =
 #if __GLASGOW_HASKELL__ >= 800
@@ -306,6 +325,62 @@ test_getDataD_kind_sig =
   True -- DataD didn't have the ability to store kind signatures prior to GHC 8.0
 #endif
 
+test_t102 :: Bool
+test_t102 =
+  $(do decs <- [d| data Foo x where MkFoo :: forall a. { unFoo :: a } -> Foo a |]
+       withLocalDeclarations decs $ do
+         [DDataD _ _ foo [DPlainTV x] _ cons _] <- dsDecs decs
+         recs <- getRecordSelectors (DConT foo `DAppT` DVarT x) cons
+         (length recs `div` 2) `eqTHSplice` 1)
+
+test_t103 :: Bool
+test_t103 =
+#if __GLASGOW_HASKELL__ >= 800
+  $(do decs <- [d| data P (a :: k) = MkP |]
+       [DDataD _ _ _ _ _ [DCon tvbs _ _ _ _] _] <- dsDecs decs
+       case tvbs of
+         [DPlainTV k, DKindedTV a (DVarT k')]
+           |  nameBase k == "k"
+           ,  nameBase a == "a"
+           ,  k == k'
+           -> [| True |]
+           |  otherwise
+           -> [| False |])
+#else
+  True -- No explicit kind variable binders prior to GHC 8.0
+#endif
+
+test_t112 :: [Bool]
+test_t112 =
+  $(do a <- newName "a"
+       b <- newName "b"
+       let [aVar, bVar] = map DVarT    [a, b]
+           [aTvb, bTvb] = map DPlainTV [a, b]
+       let fvsABExpected = [aTvb, bTvb]
+           fvsABActual   = toposortTyVarsOf [aVar, bVar]
+
+           fvsBAExpected = [bTvb, aTvb]
+           fvsBAActual   = toposortTyVarsOf [bVar, aVar]
+
+           eqAB = fvsABExpected `eqTH` fvsABActual
+           eqBA = fvsBAExpected `eqTH` fvsBAActual
+       [| [eqAB, eqBA] |])
+
+-- Unit tests for functions that compute free variables (e.g., fvDType)
+test_fvs :: [Bool]
+test_fvs =
+  $(do a <- newName "a"
+
+       let -- (Show a => Show (Maybe a)) => String
+           ty1 = DForallT
+                   []
+                   [DForallT [] [DConT ''Show `DAppT` DVarT a]
+                                (DConT ''Show `DAppT` (DConT ''Maybe `DAppT` DVarT a))]
+                   (DConT ''String)
+           b1 = fvDType ty1 `eqTH` OS.singleton a -- #93
+
+       [| [b1] |])
+
 test_kind_substitution :: [Bool]
 test_kind_substitution =
   $(do a <- newName "a"
@@ -337,10 +412,10 @@ test_kind_substitution =
            freeVars3 = fvDType substTy3
            freeVars4 = fvDType substTy4
 
-           b1 = freeVars1 `eqTH` S.singleton b
-           b2 = freeVars2 `eqTH` S.singleton b
-           b3 = freeVars3 `eqTH` S.empty
-           b4 = freeVars4 `eqTH` S.singleton k
+           b1 = freeVars1 `eqTH` OS.singleton b
+           b2 = freeVars2 `eqTH` OS.singleton b
+           b3 = freeVars3 `eqTH` OS.empty
+           b4 = freeVars4 `eqTH` OS.singleton k
        [| [b1, b2, b3, b4] |])
 
 test_lookup_value_type_names :: [Bool]
@@ -350,7 +425,7 @@ test_lookup_value_type_names =
        typeName <- newName nameStr
        let tyDec = DTySynD typeName [] (DConT ''Bool)
            decs  = decsToTH [ DLetDec (DSigD valName (DConT ''Bool))
-                            , DLetDec (DValD (DVarPa valName) (DConE 'False))
+                            , DLetDec (DValD (DVarP valName) (DConE 'False))
                             , tyDec ]
            lookupReify lookup_fun = withLocalDeclarations decs $ do
                                       Just n <- lookup_fun nameStr
@@ -369,6 +444,14 @@ local_reifications = $(do decs <- reifyDecs
                           let m_infos' = assumeStarT m_infos
                           ListE <$> mapM (Syn.lift . show) (unqualify m_infos'))
 
+type T123G = Either () ()
+type T123F = Either T123G T123G
+type T123E = Either T123F T123F
+type T123D = Either T123E T123E
+type T123C = Either T123D T123D
+type T123B = Either T123C T123C
+type T123A = Either T123B T123B
+
 $reifyDecs
 
 $(return [])  -- somehow, this is necessary to get the staging correct for the
@@ -428,7 +511,13 @@ test_matchTy =
 
      -- GHC 7.6 uses containers-0.5.0.0 which doesn't have a good Data instance
      -- for Map. So we have to convert to lists before comparing.
-    eq = geq `on` fmap M.toList
+    eq = (==) `on` fmap M.toList
+
+-- Test that type synonym expansion is efficient
+test_t123 :: ()
+test_t123 =
+  $(do _ <- expand (DConT ''T123A)
+       [| () |])
 
 main :: IO ()
 main = hspec $ do
@@ -444,10 +533,6 @@ main = hspec $ do
                                      [inst1, inst2] <- reifyInstances ''Show [ty]
                                      inst1 `eqTHSplice` inst2)
 
-#if __GLASGOW_HASKELL__ < 707
-    it "passes roles test" $ (decsToTH [ds_role_test]) `eqTH` role_test
-#endif
-
     it "makes type names" $ test_mkName
 
     it "fixes bug 8884" $ test_bug8884
@@ -466,8 +551,22 @@ main = hspec $ do
 
     it "expands type synonyms in kinds" $ test_t85
 
+    it "toposorts free variables in polytypes" $ test_t92
+
+    it "expands type synonyms in type variable binders" $ test_t97
+
+    it "collects GADT record selectors correctly" $ test_t102
+
+    it "quantifies kind variables in desugared ADT constructors" $ test_t103
+
     it "reifies data type return kinds accurately" $ test_getDataD_kind_sig
 
+    zipWithM (\b n -> it ("toposorts free variables deterministically " ++ show n) b)
+      test_t112 [1..]
+
+    zipWithM (\b n -> it ("computes free variables correctly " ++ show n) b)
+      test_fvs [1..]
+
     -- Remove map pprints here after switch to th-orphans
     zipWithM (\t t' -> it ("can do Type->DType->Type of " ++ t) $ t == t')
              $(sequence round_trip_types >>= Syn.lift . map pprint)
